import abc
import os

import yaml
from hotsos.core.config import HotSOSConfig
from hotsos.core.log import log


class YDefsLoader(object):
    """ Load yaml definitions. """

    def __init__(self, ytype):
        """
        @param ytype: the type of defs we are loading i.e. defs/<ytype>
        """
        self.ytype = ytype
        self._loaded_defs = None
        self.stats_num_files_loaded = 0

    def _is_def(self, abs_path):
        return abs_path.endswith('.yaml')

    def _get_yname(self, path):
        return os.path.basename(path).partition('.yaml')[0]

    def _get_defs_recursive(self, path):
        """ Recursively find all yaml/files beneath a directory. """
        defs = {}
        for entry in os.listdir(path):
            abs_path = os.path.join(path, entry)
            if os.path.isdir(abs_path):
                subdefs = self._get_defs_recursive(abs_path)
                defs[os.path.basename(abs_path)] = subdefs
            else:
                if not self._is_def(abs_path):
                    continue

                if self._get_yname(abs_path) == os.path.basename(path):
                    with open(abs_path) as fd:
                        log.debug("applying dir globals %s", entry)
                        defs.update(yaml.safe_load(fd.read()) or {})

                    # NOTE: these files do not count towards the total loaded
                    # since they are only supposed to contain directory-level
                    # globals that apply to other definitions in or below this
                    # directory.
                    continue

                with open(abs_path) as fd:
                    self.stats_num_files_loaded += 1
                    _content = yaml.safe_load(fd.read()) or {}
                    defs[self._get_yname(abs_path)] = _content

        return defs

    @property
    def plugin_defs(self):
        """ Load yaml defs for the current plugin and type. """
        log.debug('loading %s definitions for plugin=%s', self.ytype,
                  HotSOSConfig.plugin_name)

        if self._loaded_defs:
            return self._loaded_defs

        path = os.path.join(HotSOSConfig.plugin_yaml_defs, self.ytype,
                            HotSOSConfig.plugin_name)
        # reset
        self.stats_num_files_loaded = 0
        if os.path.isdir(path):
            loaded = self._get_defs_recursive(path)
            log.debug("YDefsLoader: plugin %s loaded %s file(s)",
                      HotSOSConfig.plugin_name, self.stats_num_files_loaded)
            # only return if we loaded actual definitions (not just globals)
            if self.stats_num_files_loaded:
                self._loaded_defs = loaded
                return loaded


class YHandlerBase(object):

    @property
    @abc.abstractmethod
    def searcher(self):
        """
        @return: FileSearcher object to be used by this handler.
        """

    @abc.abstractmethod
    def load(self):
        """
        Load definitions from yaml in preparation for them to be processed.
        """

    @abc.abstractmethod
    def run(self):
        """ Process operations. """

    @abc.abstractmethod
    def load_and_run(self):
        """ Load and run. """



from hotsos.core.plugins.system.system import SystemBase
from hotsos.core import plugintools


class SystemChecksBase(SystemBase, plugintools.PluginPartBase):
    plugin_name = 'system'
    plugin_root_index = 1

    @property
    def plugin_runnable(self):
        # Always run
        return True



from functools import cached_property

from hotsos.core.host_helpers import SnapPackageHelper
from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.requires import (
    intercept_exception,
    YRequirementTypeBase,
    PackageCheckItemsBase,
)


class SnapCheckItems(PackageCheckItemsBase):

    # NOTE: the following pylint disable can be removed when we move to newer
    #       pylint.
    @cached_property
    def packaging_helper(self):  # pylint: disable=W0236
        return SnapPackageHelper(core_snaps=self.packages_to_check)

    @cached_property
    def installed_revisions(self):
        _revisions = []
        for p in self.installed:
            _revisions.append(self.packaging_helper.get_revision(p))

        return _revisions

    @cached_property
    def installed_channels(self):
        _channels = []
        for p in self.installed:
            _channels.append(self.packaging_helper.get_channel(p))

        return _channels

    def package_info_matches(self, pkg, pkg_info):
        """
        If snap package has revisions and/or channels we check them here.

        @param pkg: name of package
        @param pkg_info: list of {min: x, max: y, channel: z}
        """
        result = False

        # Ranges must be specified as min:max
        if any((i.get('min') is None and i.get('max') is not None) or
               (i.get('max') is None and i.get('min') is not None)
                for i in pkg_info):
            raise Exception("revision ranges for snap package '{}' contain "
                            "one or more range with a missing min or max "
                            "value".format(pkg))

        _channel_to_check = None
        if any(i.get('min') is not None for i in pkg_info):
            revision = int(self.packaging_helper.get_revision(pkg))
            # compare revisions for the time being
            for item in sorted(pkg_info, key=lambda i: i['max'],
                               reverse=True):
                r_max = int(item['max'])
                r_min = int(item['min'])
                if r_min <= revision <= r_max:
                    if 'channel' in item:
                        _channel_to_check = item['channel']
                        break

                    log.debug("match snap revision within range %s:%s", r_min,
                              r_max)
                    return True

        if not all(i.get('channel') is None for i in pkg_info):
            channel = self.packaging_helper.get_channel(pkg)
            if _channel_to_check is not None:
                if _channel_to_check == channel:
                    result = True
            else:
                for item in pkg_info:
                    if item['channel'] == channel:
                        result = True
                        log.debug("match snap channel %s", channel)
                        break

        return result


class YRequirementTypeSnap(YRequirementTypeBase):
    """ Provides logic to perform checks on snap packages. """

    @classmethod
    def _override_keys(cls):
        return ['snap']

    @property
    def channel(self):
        return self.content.get('channel')

    @property
    @intercept_exception
    def _result(self):
        _result = True
        items = SnapCheckItems(self.content)
        # bail on first fail i.e. if any not installed
        if not items.not_installed:
            for pkg, info in items:
                if not info:
                    continue

                _result = items.package_info_matches(pkg, info)
                # bail at first failure
                if not _result:
                    break
        else:
            log.debug("one or more packages not installed so returning False "
                      "- %s", ', '.join(items.not_installed))
            _result = False

        log.debug('requirement check: snap(s) %s (result=%s)',
                  ', '.join(items.installed), _result)
        self.cache.set('package', ', '.join(items.installed))
        self.cache.set('revision', ', '.join(items.installed_revisions))
        self.cache.set('channel', ', '.join(items.installed_channels))
        return _result



import re

from hotsos.core.host_helpers import CLIHelper, UptimeHelper
from hotsos.core.plugins.system import SystemChecksBase


class SystemSummary(SystemChecksBase):
    summary_part_index = 0

    def __0_summary_hostname(self):
        return self.hostname

    def __1_summary_os(self):
        if self.os_release_name:
            return self.os_release_name

    def __2_summary_num_cpus(self):
        if self.num_cpus:
            return self.num_cpus

    def __3_summary_load(self):
        if UptimeHelper().loadavg:
            return UptimeHelper().loadavg

    def __4_summary_virtualisation(self):
        if self.virtualisation_type:
            return self.virtualisation_type

    def __5_summary_rootfs(self):
        df_output = CLIHelper().df()
        if df_output:
            for line in df_output:
                ret = re.compile(r"(.+\/$)").match(line)
                if ret:
                    return ret[1]

    def __6_summary_unattended_upgrades(self):
        if self.unattended_upgrades_enabled:
            return "ENABLED"
        return "disabled"

    def __7_summary_date(self):
        return self.date

    def __8_summary_ubuntu_pro(self):
        return self.ubuntu_pro_status

    def __9_summary_uptime(self):
        return self.uptime



import glob
import os
import re
from functools import cached_property

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import CLIHelper, CLIHelperFile
from hotsos.core.host_helpers.config import ConfigBase
from hotsos.core.plugins.storage import StorageBase
from hotsos.core.search import (
    FileSearcher,
    SequenceSearchDef,
    SearchDef
)
from hotsos.core.utils import sort_suffixed_integers


class BcacheConfig(ConfigBase):

    def get(self, key):
        cfg = os.path.join(self.path, key)
        if os.path.exists(cfg):
            with open(cfg) as fd:
                return fd.read().strip()


class BDev(object):

    def __init__(self, path, cache):
        self.cache = cache
        self.path = path

    @property
    def cfg(self):
        return BcacheConfig(self.path)

    @property
    def dev_to_dname(self):
        for line in CLIHelper().udevadm_info_dev(device=self.dev):
            expr = r'.+\s+disk/by-dname/(.+)'
            ret = re.compile(expr).match(line)
            if ret:
                return ret[1]

    @property
    def dev(self):
        return os.path.basename(os.path.realpath(os.path.join(self.path,
                                                              'dev')))

    @cached_property
    def name(self):
        return os.path.basename(self.path)

    def __getattr__(self, key):
        cfg = os.path.join(self.path, key)
        if os.path.exists(cfg):
            with open(cfg) as fd:
                return fd.read().strip()

        raise AttributeError("{} not found in bdev config".format(key))


class Cacheset(object):

    def __init__(self, path):
        self.path = path
        self.bdevs = []

        self.uuid = os.path.basename(self.path)
        for bdev in glob.glob(os.path.join(self.path, 'bdev*')):
            self.bdevs.append(BDev(bdev, self))

    @cached_property
    def cfg(self):
        return BcacheConfig(self.path)

    def __getattr__(self, key):
        cfg = os.path.join(self.path, key)
        if os.path.exists(cfg):
            with open(cfg) as fd:
                return fd.read().strip()

        raise AttributeError("{} not found in cacheset config".format(key))


class BcacheBase(StorageBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cachesets = []

        for entry in glob.glob(os.path.join(HotSOSConfig.data_root,
                               'sys/fs/bcache/*')):
            if not os.path.isdir(entry):
                continue

            if not os.path.exists(os.path.join(entry,
                                               'cache_available_percent')):
                continue

            self.cachesets.append(Cacheset(entry))

    @cached_property
    def bcache_enabled(self):
        """ Return True if there are any backing devices configured. """
        if self.cachesets:
            for cset in self.cachesets:
                if cset.bdevs:
                    return True

        return False

    @cached_property
    def udev_bcache_devs(self):
        """ If bcache devices exist fetch information and return as a list. """
        devs = []
        with CLIHelperFile() as cli:
            s = FileSearcher()
            sdef = SequenceSearchDef(start=SearchDef(r"^P: .+/(bcache\S+)"),
                                     body=SearchDef(r"^S: disk/by-uuid/(\S+)"),
                                     tag="bcacheinfo")
            s.add(sdef, cli.udevadm_info_exportdb())
            results = s.run()
            for section in results.find_sequence_sections(sdef).values():
                dev = {}
                for r in section:
                    if r.tag == sdef.start_tag:
                        dev["name"] = r.get(1)
                    else:
                        dev["by-uuid"] = r.get(1)

                devs.append(dev)

        return devs

    def resolve_bdev_from_dev(self, devpath):
        """
        Given a device path, resolve it to a corresponding bcache BDev object.
        """
        bcache_name = None
        if 'bcache' in devpath:
            bcache_name = os.path.basename(devpath)
        else:
            ret = re.compile(r"/dev/mapper/crypt-(\S+)").search(devpath)
            if ret:
                for udev_dev in self.udev_bcache_devs:
                    if 'name' not in udev_dev:
                        continue

                    _uuid = udev_dev.get("by-uuid")
                    if not _uuid or _uuid != ret.group(1):
                        continue

                    bcache_name = udev_dev['name']

        if bcache_name:
            for cset in self.cachesets:
                for bdev in cset.bdevs:
                    if bdev.dev == bcache_name:
                        return bdev

    def is_bcache_device(self, dev):
        """
        Returns True if the device either is or is backed by a bcache device
        e.g. dmcrypt device using bcache dev.
        """
        if dev.startswith("bcache"):
            return True

        if dev.startswith("/dev/bcache"):
            return True

        if self.resolve_bdev_from_dev(dev):
            return True

        return False


class BDevsInfo(BcacheBase):

    def _get_parameter(self, key):
        all_values = []
        for cset in self.cachesets:
            for bdev in cset.bdevs:
                all_values.append(bdev.cfg.get(key))

        return all_values

    @property
    def sequential_cutoff(self):
        """
        @return: list of <str> sequential_cutoff from each bdev sorted in
                 descending order. values can be int or float with our without
                 a suffix e.g. '2' or '0.0k'.
        """
        ret = self._get_parameter("sequential_cutoff")
        return sort_suffixed_integers(ret, reverse=True)

    @property
    def cache_mode(self):
        """
        @return: list of <str> cache_mode from each bdev sorted so that the
                 settings with writeback mode selected are at the end of the
                 list.
        """
        ret = self._get_parameter("cache_mode")
        # send [writeback] to the end of the list
        return sorted(ret,
                      key='writethrough [writeback] writearound none'.__eq__)

    @property
    def writeback_percent(self):
        """
        @return: list of <int> writeback_percent from each bdev sorted in
                 ascending order.
        """
        ret = self._get_parameter("writeback_percent")
        # convert str to int
        return sorted(list(map(int, ret)))


class CachesetsInfo(BcacheBase):

    def _get_parameter(self, key):
        return [cset.cfg.get(key) for cset in self.cachesets]

    @property
    def congested_read_threshold_us(self):
        """
        @return: list of <int> congested_read_threshold_us from each cacheset
                 sorted in descending order.
        """
        ret = self._get_parameter("congested_read_threshold_us")
        # convert str to int
        return sorted(list(map(int, ret)), reverse=True)

    @property
    def congested_write_threshold_us(self):
        """
        @return: list of <int> congested_write_threshold_us from each cacheset
                 sorted in descending order.
        """
        ret = self._get_parameter("congested_write_threshold_us")
        # convert str to int
        return sorted(list(map(int, ret)), reverse=True)

    @property
    def cache_available_percent(self):
        """
        @return: list of <int> cache_available_percent from each cacheset
                 sorted in ascending order.
        """
        ret = self._get_parameter("cache_available_percent")
        # convert str to int
        return sorted(list(map(int, ret)))


class BcacheChecksBase(BcacheBase):

    @property
    def summary_subkey(self):
        return 'bcache'

    @property
    def plugin_runnable(self):
        return self.bcache_enabled



import datetime

import yaml
from hotsos.core.analytics import LogEventStats, SearchResultIndices
from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import CLIHelper
from hotsos.core.issues import (
    IssueContext,
    IssuesManager,
    NeutronL3HAWarning,
    OpenstackWarning
)
from hotsos.core.log import log
from hotsos.core.plugins.openstack.common import (
    OpenstackChecksBase,
    OpenstackEventHandlerBase,
    OpenstackEventCallbackBase,
)
from hotsos.core.plugins.openstack.neutron import NeutronHAInfo
from hotsos.core.search import (
    FileSearcher,
    SearchConstraintSearchSince,
)
from hotsos.core import utils
from hotsos.core.utils import sorted_dict
from hotsos.core.ycheck.engine.properties.search import CommonTimestampMatcher

VRRP_TRANSITION_WARN_THRESHOLD = 8


class ApacheEventCallback(OpenstackEventCallbackBase):
    event_group = 'apache'
    event_names = ['connection-refused']

    def __call__(self, event):
        ports_max = {}
        context = {}

        results = []
        for result in event.results:
            # save some context info
            path = event.searcher.resolve_source_id(result.source_id)
            if path in context:
                context[path].append(result.linenumber)
            else:
                context[path] = [result.linenumber]

            month = datetime.datetime.strptime(result.get(1), '%b').month
            day = result.get(2)
            year = result.get(3)
            addr = result.get(4)
            port = result.get(5)
            results.append({'date': "{}-{}-{}".format(year, month, day),
                            'key': "{}:{}".format(addr, port)})

        conns_refused = self.categorise_events(event, results=results)
        for addrs in conns_refused.values():
            for addr, count in addrs.items():
                port = addr.partition(':')[2]
                # allow a small number of connection refused errors on a given
                # day
                if count < 5:
                    continue

                if port not in ports_max:
                    ports_max[port] = count
                else:
                    ports_max[port] = max(count, ports_max[port])

        if ports_max:
            msg = ('apache is reporting connection refused errors for the '
                   'following ports which could mean some services are not '
                   'working properly - {} - please check.'.
                   format(','.join(ports_max.keys())))
            IssuesManager().add(OpenstackWarning(msg), IssueContext(**context))

        return sorted_dict(conns_refused)


class ApacheEventChecks(OpenstackEventHandlerBase):
    event_group = 'apache'

    def __104_summary_apache(self):
        return self.final_event_results


class APIEventsCallback(OpenstackEventCallbackBase):
    event_group = 'http-requests'
    event_names = ['neutron']

    def __call__(self, event):
        results = [{'date': r.get(1),
                    'key': r.get(2)} for r in event.results]
        ret = self.categorise_events(event, results=results)
        if ret:
            return ret


class APIEvents(OpenstackEventHandlerBase):
    event_group = 'http-requests'

    def __105_summary_http_requests(self):
        out = self.final_event_results
        if out:
            return out


class AgentEventsCallback(OpenstackEventCallbackBase):
    event_group = 'neutron.agents'
    ovsdbapp_event_names = ['ovsdbapp-nb-leader-reconnect',
                            'ovsdbapp-sb-leader-reconnect']
    ovn_mech_driver_events = ['ovn-resource-revision-bump',
                              'ovsdb-monitor-router-binding-transitions',
                              'ovsdb-transaction-aborted']
    event_names = ['rpc-loop', 'router-spawn-events', 'router-updates']
    event_names += ovsdbapp_event_names + ovn_mech_driver_events

    def _get_event_stats(self, results, tag_prefix, custom_idxs=None):
        stats = LogEventStats(results, tag_prefix, custom_idxs=custom_idxs)
        stats.run()
        top5 = stats.get_top_n_events_sorted(5)
        if not top5:
            return

        return {"top": top5,
                "stats": stats.get_event_stats()}

    def __call__(self, event):
        agent = event.section
        if event.name == 'ovsdb-monitor-router-binding-transitions':
            # The ovsdbmonitor will print bindings which will eventually
            # reflect transitions so we need to filter out contiguous bindings
            # to the same chassis so that we are just left with transitions
            # that we can then count.
            results = []
            for r in event.results:
                results.append({'date': r.get(1), 'time': r.get(2),
                                'router': r.get(3), 'key': r.get(4)})

            results = sorted(results,
                             key=lambda e:
                             datetime.datetime.strptime(
                                 "{} {}".format(e['date'],
                                                e['time']),
                                 '%Y-%m-%d %H:%M:%S'))

            new_results = []
            last_known = {}
            host_key = 'key'
            for item in results:
                router = item['router']
                if router not in last_known:
                    last_known[router] = item
                elif last_known[router][host_key] != item[host_key]:
                    # remember hosts port transitioned from
                    new_results.append(last_known[router])
                    last_known[router] = item

            ret = self.categorise_events(event, results=new_results)
            if ret:
                return {event.name: ret}, agent
        elif event.name in self.ovsdbapp_event_names + \
                self.ovn_mech_driver_events:
            if event.name == 'ovn-resource-revision-bump':
                ret = self.categorise_events(event, max_results_per_date=5)
            elif event.name == 'ovsdb-transaction-aborted':
                ret = {}
                for result in event.results:
                    _date = result.get(1)
                    if _date in ret:
                        ret[_date] += 1
                    else:
                        ret[_date] = 1

                ret = sorted_dict(ret)
            else:
                ret = self.categorise_events(event)

            if ret:
                return {event.name: ret}, agent
        elif event.name in ['rpc-loop', 'router-spawn-events']:
            ret = self._get_event_stats(event.results, event.search_tag)
            if ret:
                return {event.name: ret}, agent
        else:
            sri = SearchResultIndices(event_id_idx=4,
                                      metadata_idx=3,
                                      metadata_key='router')
            ret = self._get_event_stats(event.results, event.search_tag,
                                        custom_idxs=sri)
            if ret:
                return {event.name: ret}, agent


class NeutronAgentEventChecks(OpenstackEventHandlerBase):
    """
    Loads events we want to check from definitions yaml and executes them. The
    results are sorted by date and the "top 5" are presented along with stats
    on the full set of samples.
    """
    event_group = 'neutron.agents'

    # NOTE: can share summary index with l3 agent since they will never be
    #       colocated.
    def __102_summary_neutron_server(self):
        out = self.final_event_results or {}
        return out.get('neutron-server')

    def __102_summary_neutron_l3_agent(self):
        out = self.final_event_results or {}
        return out.get('neutron-l3-agent')

    def __103_summary_neutron_ovs_agent(self):
        out = self.final_event_results or {}
        return out.get('neutron-ovs-agent')


class OctaviaAgentEventsCallback(OpenstackEventCallbackBase):
    event_group = 'octavia'
    event_names = ['lb-failover-auto', 'lb-failover-manual',
                   'amp-missed-heartbeats']

    def __call__(self, event):
        if event.name in ['lb-failover-auto', 'lb-failover-manual']:
            results = []
            for e in event.results:
                payload = yaml.safe_load(e.get(3))
                lb_id = payload.get('load_balancer_id')
                if lb_id is None:
                    continue

                results.append({'date': e.get(1), 'time': e.get(2),
                                'key': lb_id})

            ret = self.categorise_events(event, results=results,
                                         key_by_date=False)
            if ret:
                failover_type = event.name.rpartition('-')[2]
                return {failover_type: ret}, 'lb-failovers'
        else:
            missed_heartbeats = self.categorise_events(event,
                                                       key_by_date=False)
            if not missed_heartbeats:
                return

            # sort each amp by occurences
            for ts_date, amps in missed_heartbeats.items():
                amps_sorted = utils.sorted_dict(amps, key=lambda e: e[1],
                                                reverse=True)
                missed_heartbeats[ts_date] = amps_sorted

            # then sort by date
            return utils.sorted_dict(missed_heartbeats)


class OctaviaAgentEventChecks(OpenstackEventHandlerBase):
    event_group = 'octavia'

    def __106_summary_octavia(self):
        return self.final_event_results


class PCINotFoundCallback(OpenstackEventCallbackBase):
    event_group = 'nova.nova-compute'
    event_names = ['pci-dev-not-found']

    def __call__(self, event):
        ret = self.categorise_events(event)
        if ret:
            return ret, 'PciDeviceNotFoundById'


class NovaComputeEventChecks(OpenstackEventHandlerBase):
    event_group = 'nova.nova-compute'

    def __107_summary_nova(self):
        return self.final_event_results


class ApparmorCallback(OpenstackEventCallbackBase):
    event_group = 'apparmor'
    event_names = ['nova', 'neutron']

    def __call__(self, event):
        results = [{'date': "{} {}".format(r.get(1), r.get(2)),
                    'time': r.get(3),
                    'key': r.get(4)} for r in event.results]
        ret = self.categorise_events(event, results=results,
                                     key_by_date=False)
        if ret:
            # event.name must be the service name, event.section is the aa
            # action.
            return {event.name: ret}, event.section


class AgentApparmorChecks(OpenstackEventHandlerBase):
    event_group = 'apparmor'

    def __108_summary_apparmor(self):
        return self.final_event_results


class L3HACallback(OpenstackEventCallbackBase):
    event_group = 'neutron.ml2-routers'
    event_names = ['vrrp-transitions']

    def check_vrrp_transitions(self, transitions):
        # there will likely be a large number of transitions if we look across
        # all time so dont run this check.
        if HotSOSConfig.use_all_logs:
            return

        max_transitions = 0
        warn_count = 0
        threshold = VRRP_TRANSITION_WARN_THRESHOLD
        for router in transitions:
            r = transitions[router]
            _transitions = sum(t for d, t in r.items())
            if _transitions > threshold:
                max_transitions = max(_transitions, max_transitions)
                warn_count += 1

        if warn_count:
            msg = ("{} router(s) have had more than {} vrrp transitions "
                   "(max={}) in the last 24 hours.".format(warn_count,
                                                           threshold,
                                                           max_transitions))
            IssuesManager().add(NeutronL3HAWarning(msg))

    def __call__(self, event):
        results = []
        for r in event.results:
            router = NeutronHAInfo().find_router_with_vr_id(r.get(3))
            if not router:
                log.debug("could not find router with vr_id %s", r.get(3))
                continue

            results.append({'date': r.get(1), 'time': r.get(2),
                            'key': router.uuid})

        transitions = self.categorise_events(event, results=results,
                                             key_by_date=False)
        if transitions:
            # run checks
            self.check_vrrp_transitions(transitions)
            # add info to summary
            return {'transitions': transitions}, 'keepalived'


class NeutronL3HAEventChecks(OpenstackEventHandlerBase):
    event_group = 'neutron.ml2-routers'

    def journalctl_args(self):
        """ Args callback for event cli command """
        args = []
        kwargs = {'unit': 'neutron-l3-agent'}
        if not HotSOSConfig.use_all_logs:
            kwargs['date'] = CLIHelper().date(format="--iso-8601")

        return args, kwargs

    def __109_summary_neutron_l3ha(self):
        return self.final_event_results


class AgentEventChecks(OpenstackChecksBase):
    summary_part_index = 7

    def _run_checks(self, checks):
        # Only run if we think Openstack is installed.
        if not self.openstack_installed:
            return

        searcher = FileSearcher(constraint=SearchConstraintSearchSince(
                                        ts_matcher_cls=CommonTimestampMatcher))
        check_objs = [c(searcher=searcher) for c in checks]
        for check in check_objs:
            check.load()

        results = searcher.run()
        _final_results = {}
        for check in check_objs:
            check.run(results)
            check_results = check.raw_output
            if check_results:
                _final_results.update(check_results)

        return _final_results

    def __100_summary_api_info(self):
        checks = [APIEvents]
        results = self._run_checks(checks)
        if results:
            return results

    def __101_summary_agent_checks(self):
        checks = [AgentApparmorChecks,
                  ApacheEventChecks,
                  NeutronL3HAEventChecks,
                  NeutronAgentEventChecks,
                  NovaComputeEventChecks,
                  OctaviaAgentEventChecks]
        results = self._run_checks(checks)
        if results:
            return results



#!/usr/bin/python3
import logging
import os
import tempfile
from functools import cached_property

from hotsos.core.config import HotSOSConfig

log = logging.getLogger('hotsos')


class LoggingManager(object):

    def __init__(self):
        self.delete_temp_file = True

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args, **kwargs):
        self.stop()

        # Allow raise
        return False

    @property
    def _format(self):
        return ("%(asctime)s.%(msecs)03d %(process)d %(levelname)s %(name)s "
                "[-] %(message)s")

    @cached_property
    def _handler(self):
        if HotSOSConfig.debug_mode:
            return logging.StreamHandler()

        return logging.FileHandler(self.temp_log_path)

    @cached_property
    def temp_log_path(self):
        return tempfile.mktemp(suffix='hotsos.log')

    def setup_deps_loggers(self, level):
        # Set logging level for dependencies
        logging.getLogger('searchkit').setLevel(level=level)
        logging.getLogger('searchkit').addHandler(self._handler)
        # NOTE: dont set log level here as it is controlled using env var
        logging.getLogger('propertree').addHandler(self._handler)

    def start(self, level=logging.DEBUG):
        log.setLevel(level)
        if log.hasHandlers():
            return

        self._handler.setFormatter(logging.Formatter(self._format))
        log.addHandler(self._handler)
        self.setup_deps_loggers(level)

    def stop(self):
        if os.path.exists(self.temp_log_path) and self.delete_temp_file:
            log.debug("removing temporary log file %s", self.temp_log_path)
            os.remove(self.temp_log_path)



from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.common import (
    YPropertyMappedOverrideBase,
    LogicalCollectionHandler,
    add_to_property_catalog,
)
from hotsos.core.ycheck.engine.properties.requires.types import (
    apt,
    snap,
    config,
    pebble,
    systemd,
    property as rproperty,
    path,
    varops,
)


class YPropertyRequiresBase(YPropertyMappedOverrideBase,
                            LogicalCollectionHandler):
    CACHE_CHECK_KEY = '__PREVIOUSLY_CACHED_PROPERTY_TYPE'

    @classmethod
    def _override_mapped_member_types(cls):
        return [apt.YRequirementTypeAPT,
                snap.YRequirementTypeSnap,
                config.YRequirementTypeConfig,
                pebble.YRequirementTypePebble,
                systemd.YRequirementTypeSystemd,
                rproperty.YRequirementTypeProperty,
                path.YRequirementTypePath,
                varops.YPropertyVarOps]

    def get_item_result_callback(self, item, grouped=False):
        """
        Evaluate item and return its result.

        Also copy the cache of the item into the local cache. If part of a
        logical group whereby more than one item is being evaluated to get
        a final results, we only copy the cache data if the result is True
        and  either (a) it is the first to be saved or (b) it has the same type
        as the previous one to be saved.

        For example with the following group:

        or:
          - apt: foo
          - apt: bar

        If package foo exists and bar does not, the final contents of the group
        will that from 'apt: foo'. If both exists then it will be the result of
        'apt: bar'.

        @param item: YRequirementType object
        @param grouped: If True, item is part of a (logical) group of one or
                        more items or a list with more than one item to which
                        to default logical operator (AND) will be applied.
        @return: boolean result i.e. True or False
        """
        result = item()
        previously_cached = getattr(self.cache, self.CACHE_CHECK_KEY)
        req_prop_type = item.__class__.__name__
        if not grouped or (result and
                           previously_cached in (None, req_prop_type)):
            if previously_cached is None:
                self.cache.set(self.CACHE_CHECK_KEY, req_prop_type)

            log.debug("merging item (type=%s) cache id=%s into cache id=%s",
                      req_prop_type, item.cache.id, self.cache.id)
            self.cache.merge(item.cache)

        return result

    @property
    def passes(self):
        """
        Content can either be a single requirement, dict of requirement groups
        or list of requirements. List may contain individual requirements or
        groups.
        """
        log.debug("running requirement")
        try:
            result = self.run_collection()
        except Exception:
            log.exception("exception caught during run_collection:")
            raise

        self.cache.set('passes', result)
        return result


class YPropertyRequiresLogicalGroupsExtension(YPropertyRequiresBase):

    @classmethod
    def _override_keys(cls):
        return LogicalCollectionHandler.VALID_GROUP_KEYS


@add_to_property_catalog
class YPropertyRequires(YPropertyRequiresBase):

    @classmethod
    def _override_keys(cls):
        return ['requires']

    @classmethod
    def _override_mapped_member_types(cls):
        return super()._override_mapped_member_types() + \
                    [YPropertyRequiresLogicalGroupsExtension]



import os
import re
from collections import UserDict

from hotsos.core.config import HotSOSConfig
from hotsos.core.log import log
from hotsos.core.plugins.openstack.common import OpenstackChecksBase
from hotsos.core.search import (
    FileSearcher,
    SearchDef,
    SearchConstraintSearchSince,
)
from hotsos.core.ycheck.engine.properties.search import CommonTimestampMatcher


class AgentExceptionCheckResults(UserDict):

    def __init__(self, results, source_id_resolver_callback):
        """
        @param results: list of searchkit.SearchResult objects grouped by
                        agent/log in which they were found.
        @param source_id_resolver_callback: searchkit results contain a
                                            reference to the path that matched
                                            the result and this callback is
                                            called to resolve back to a path.
        """
        self.results = results
        self.source_id_resolver_callback = source_id_resolver_callback
        self.data = {}
        for name, _results in self.results.items():
            self.data[name] = self._tally_results(_results)

    @property
    def agents(self):
        """ Returns a list of agents for that have raised exceptions. """
        return list(self.results.keys())

    def _tally_results(self, results):
        """ Tally search results.

        Returns a dictionary with results grouped/presented by date but
        can optionally be grouped by time for more granularity.

        Each search result is expected to have the following match groups:
            1: date
            2: time
            3: log entry
        """
        exceptions = {}
        for result in results:
            # strip leading/trailing quotes
            exc_name = result.get(3).strip("'")
            if exc_name not in exceptions:
                exceptions[exc_name] = {}

            # results are grouped by date or datetime
            ts_date = result.get(1)
            if HotSOSConfig.event_tally_granularity == 'time':
                # use hours and minutes only
                ts_time = re.compile(r'(\d+:\d+).+').search(result.get(2))[1]
                key = "{}_{}".format(ts_date, ts_time)
            else:
                key = str(ts_date)

            if key not in exceptions[exc_name]:
                exceptions[exc_name][key] = 0

            exceptions[exc_name][key] += 1

        if not exceptions:
            return

        for exc_type in exceptions:
            exceptions_sorted = {}
            for k, v in sorted(exceptions[exc_type].items(),
                               key=lambda x: x[0]):
                exceptions_sorted[k] = v

            exceptions[exc_type] = exceptions_sorted

        return exceptions

    @property
    def exceptions_raised(self):
        """ Return a list of exceptions raised by this agent. """
        _exceptions = set()
        for results in self.values():
            for exception in results:
                _exceptions.add(exception)

        return list(_exceptions)

    @property
    def files_w_exceptions(self):
        """ Return a list of files containing exceptions. """
        files = []
        for results in self.results.values():
            sources = set(r.source_id for r in results)
            files.extend([self.source_id_resolver_callback(s)
                          for s in sources])

        return files


class AgentExceptionChecks(OpenstackChecksBase):
    """
    Openstack services/agents will log exceptions using ERROR and
    WARNING log levels depending on who raised them and their
    importance. This class provides a way to identify exceptions in
    logs and provide a per-agent tally by date or time.
    """
    summary_part_index = 6

    def __init__(self):
        super().__init__()
        self._agent_results = None
        c = SearchConstraintSearchSince(
                                      ts_matcher_cls=CommonTimestampMatcher)
        self.searchobj = FileSearcher(constraint=c)
        # The following are expected to be logged using WARNING log level.
        self._agent_warnings = {
            'nova': ['MessagingTimeout',
                     'DiskNotFound',
                     r"Timeout waiting for \[\('\S+',",
                     ],
            'neutron': [r'OVS is dead',
                        r'MessagingTimeout',
                        ]
        }

        # The following are expected to be ERROR. This is typically used to
        # catch events that are not logged as an exception with the usual
        # Traceback format.
        self._agent_errors = {
            'neutron': [r'RuntimeError'],
            'keystone': [r'\([a-zA-Z\.]+\)']
        }

    def _add_agent_searches(self, project, agent_name, logs_path,
                            expr_template):
        """
        Add searches we want to apply to agent.

        @param project: OSTProject object
        @param agent_name: name of agent
        @param logs_path: path to logs we want to search
        @param expr_template: generic search template we use to search all/any
                              exception types in any log file.
        """
        constraints = True
        # keystone logs have cruft at the start of each line and won't be
        # verifiable with the standard log expr so just disable constraints
        # for these logs for now.
        if project.name == 'keystone':
            constraints = False

        tag = "{}.{}".format(project.name, agent_name)
        if project.exceptions:
            exc_names = "(?:{})".format('|'.join(project.exceptions))
            expr = expr_template.format(exc_names)
            self.searchobj.add(SearchDef(expr, tag=tag + '.error',
                                         hint='( ERROR | Traceback)'),
                               logs_path,
                               allow_global_constraints=constraints)

        warn_exprs = self._agent_warnings.get(project.name, [])
        if warn_exprs:
            values = "(?:{})".format('|'.join(warn_exprs))
            expr = expr_template.format(values)
            self.searchobj.add(SearchDef(expr, tag=tag + '.warning',
                                         hint='WARNING'), logs_path,
                               allow_global_constraints=constraints)

        err_exprs = self._agent_errors.get(project.name, [])
        if err_exprs:
            values = "(?:{})".format('|'.join(err_exprs))
            expr = expr_template.format(values)
            sd = SearchDef(expr, tag=tag + '.error', hint='ERROR')
            self.searchobj.add(sd, logs_path,
                               allow_global_constraints=constraints)

    def _load(self):
        """Register searches for exceptions as well as any other type of issue
        we might want to catch like warnings etc which may not be errors or
        exceptions.
        """
        log.debug("loading exception search defs")
        for project in self.ost_projects.all.values():
            if not project.installed:
                log.debug("%s is not installed - excluding from exception "
                          "checks", project.name)
                continue

            log.debug("%s is installed so adding to searches", project.name)

            wsgi_prefix = ''
            if 'apache2' in project.services:
                # NOTE: services running under apache may have their logs (e.g.
                # barbican-api.log) prepended with apache/mod_wsgi info so do
                # this way to account for both. If present, the prefix will be
                # ignored and not count towards the result.
                wsgi_prefix = r'\[[\w :\.]+\].+\]\s+'

            keystone_prefix = ''
            if project.name == 'keystone':
                # keystone logs contain the (module_name): at the beginning of
                # the line.
                keystone_prefix = r'\(\S+\):\s+'

            prefix_match = ''
            if all([wsgi_prefix, keystone_prefix]):
                prefix_match = r'(?:{}|{})?'.format(wsgi_prefix,
                                                    keystone_prefix)
            elif any([wsgi_prefix, keystone_prefix]):
                prefix_match = (r'(?:{})?'.
                                format(wsgi_prefix or keystone_prefix))

            # Sometimes the exception is printed with just the class name
            # and sometimes it is printed with a full import path e.g.
            # MyExc or somemod.MyExc so we need to account for both.
            exc_obj_full_path_match = r'(?:\S+\.)?'
            expr_template = (r"^{}([0-9\-]+) (\S+) .+\S+\s({}{{}})[\s:\.]".
                             format(prefix_match, exc_obj_full_path_match))

            # NOTE: don't check exceptions for deprecated services
            for agent, log_paths in project.log_paths(
                    include_deprecated_services=False):
                for path in log_paths:
                    path = os.path.join(HotSOSConfig.data_root, path)
                    if HotSOSConfig.use_all_logs:
                        path = "{}*".format(path)

                    self._add_agent_searches(project, agent, path,
                                             expr_template)

    def _run(self, search_results):
        """ Process search results to see if we got any hits.

        @param search_results: a searchkit.SearchResultsCollection object.
        @return: a dictionary of services and underlying agents with any
                 exceptions they have raised.
        """
        log.debug("processing exception search results")
        agent_exceptions = {}
        for name, project in self.ost_projects.all.items():
            if not project.installed:
                continue

            for log_level in ['warning', 'error']:
                agent_results = {}
                for agent in project.services:
                    tag = "{}.{}".format(name, agent)
                    results = search_results.find_by_tag(tag + '.' + log_level)
                    if results:
                        agent_results[agent] = results

                if not agent_results:
                    continue

                if log_level not in agent_exceptions:
                    agent_exceptions[log_level] = {}

                callback = self.searchobj.catalog.source_id_to_path
                _results = AgentExceptionCheckResults(agent_results, callback)
                agent_exceptions[log_level][name] = _results

        return agent_exceptions

    @property
    def agent_results(self):
        if self._agent_results is not None:
            return self._agent_results

        self._load()
        self._agent_results = self._run(self.searchobj.run())
        return self._agent_results

    def __200_summary_agent_exceptions(self):
        """
        Only ERROR level exceptions
        """
        if self.agent_results and 'error' in self.agent_results:
            _exc_info = {}
            for svc, results in self.agent_results['error'].items():
                _exc_info[svc] = dict(results)

            return {agent: dict(info) for agent, info in _exc_info.items()}

    def __201_summary_agent_warnings(self):
        """
        Only WARNING level exceptions
        """
        if self.agent_results and 'warning' in self.agent_results:
            _exc_info = {}
            for svc, results in self.agent_results['warning'].items():
                _exc_info[svc] = dict(results)

            return {agent: dict(info) for agent, info in _exc_info.items()}



import re

from hotsos.core.issues import IssuesManager, CephOSDError
from hotsos.core.plugins.storage.ceph import (
    CephChecksBase,
    CephEventCallbackBase,
)
from hotsos.core.ycheck.events import EventHandlerBase

CEPH_ID_FROM_LOG_PATH_EXPR = r'.+ceph-osd\.(\d+)\.log'


class EventCallbackMisc(CephEventCallbackBase):
    event_group = 'ceph'
    event_names = ['heartbeat-no-reply', 'osd-reported-failed',
                   'mon-elections-called', 'superblock-read-error']

    def __call__(self, event):
        if event.name == 'superblock-read-error':
            msg = ('Detected superblock read errors which indicates an OSD '
                   'disk failure or its likely failure in the near future. '
                   'This drive needs to be inspected further using '
                   'sar/smartctl.')
            IssuesManager().add(CephOSDError(msg))
            return

        key_by_date = event.name == 'heartbeat-no-reply'
        return self.categorise_events(event, key_by_date=key_by_date)


class EventCallbackSlowRequests(CephEventCallbackBase):
    event_group = 'ceph'
    event_names = ['slow-requests']

    def __call__(self, event):
        slow_requests = {}
        for result in sorted(event.results, key=lambda r: r.get(1)):
            date = result.get(1)
            count = result.get(2)
            if date not in slow_requests:
                slow_requests[date] = int(count)
            else:
                slow_requests[date] += int(count)

        return slow_requests


class EventCallbackCRCErrors(CephEventCallbackBase):
    event_group = 'ceph'
    event_names = ['crc-err-bluestore', 'crc-err-rocksdb']

    def __call__(self, event):
        osd_type = event.name.rpartition('_')[2]
        c_expr = re.compile(CEPH_ID_FROM_LOG_PATH_EXPR)
        results = []
        for r in event.results:
            ret = c_expr.match(event.searcher.resolve_source_id(r.source_id))
            if ret:
                key = "osd.{}".format(ret.group(1))
            else:
                key = None

            results.append({'date': r.get(1), 'key': key})

        ret = self.categorise_events(event, results=results,
                                     squash_if_none_keys=True)

        # If on any particular day there were > 3 crc errors for a
        # particular osd we raise an issue since that indicates they are
        # likely to reflect a real problem.
        osds_in_err = set()
        osd_err_max = 0
        # ret is keyed by day
        for osds in ret.values():
            # If we were unable to glean the osd id from the search results
            # this will not be a dict so skip.
            if type(osds) != dict:
                continue

            for osd, num_errs in osds.items():
                if num_errs > 3:
                    if num_errs > osd_err_max:
                        osd_err_max = num_errs

                    osds_in_err.add(osd)

        if osds_in_err:
            msg = ("{} osds ({}) found with > 3 {} crc errors (max={}) "
                   "each within a 24hr period - please investigate".
                   format(len(osds_in_err), ','.join(osds_in_err),
                          osd_type, osd_err_max))
            IssuesManager().add(CephOSDError(msg))

        return ret


class EventCallbackHeartbeatPings(CephEventCallbackBase):
    event_group = 'ceph'
    event_names = ['long-heartbeat-pings']

    def __call__(self, event):
        c_expr = re.compile(CEPH_ID_FROM_LOG_PATH_EXPR)
        results = []
        for r in event.results:
            ret = c_expr.match(event.searcher.resolve_source_id(r.source_id))
            if ret:
                key = "osd.{}".format(ret.group(1))
            else:
                key = None

            results.append({'date': r.get(1), 'key': key})

        return self.categorise_events(event, results=results,
                                      squash_if_none_keys=True)


class CephEventHandler(CephChecksBase, EventHandlerBase):
    event_group = 'ceph'
    summary_part_index = 1

    @property
    def summary(self):
        # mainline all results into summary root
        return self.load_and_run()



import abc
import os
import re

from hotsos.core.log import log


class ConfigBase(abc.ABC):

    def __init__(self, path):
        self.path = path

    @classmethod
    def squash_int_range(cls, ilist):
        """Takes a list of integers and squashes consecutive values into a
        string range. Returned list contains mix of strings and ints.
        """
        irange = []
        rstart = None
        rprev = None

        sorted(ilist)
        for i, value in enumerate(ilist):
            if rstart is None:
                if i == (len(ilist) - 1):
                    irange.append(value)
                    break

                rstart = value

            if rprev is not None:
                if rprev != (value - 1):
                    if rstart == rprev:
                        irange.append(rstart)
                    else:
                        irange.append("{}-{}".format(rstart, rprev))
                        if i == (len(ilist) - 1):
                            irange.append(value)

                    rstart = value
                elif i == (len(ilist) - 1):
                    irange.append("{}-{}".format(rstart, value))
                    break

            rprev = value

        return ','.join(irange)

    @classmethod
    def expand_value_ranges(cls, ranges):
        """
        Takes a string containing ranges of values such as 1-3 and 4,5,6,7 and
        expands them into a single list.
        """
        if not ranges:
            return ranges

        expanded = []
        ranges = ranges.split(',')
        for subrange in ranges:
            # expand ranges
            subrange = subrange.partition('-')
            if subrange[1] == '-':
                expanded += range(int(subrange[0]), int(subrange[2]) + 1)
            else:
                for val in subrange[0].split():
                    expanded.append(int(val))

        return sorted(expanded)

    @property
    def exists(self):
        if os.path.exists(self.path):
            return True

        return False

    @abc.abstractmethod
    def get(self, key, section=None, expand_to_list=False):
        """ Get a config value. """


class SectionalConfigBase(ConfigBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sections = {}
        # this provides an easy sectionless lookup but is prone to collisions.
        # always returns the last value for key found in config file.
        self._flattened_config = {}
        self._load()

    @staticmethod
    def bool_str(val):
        if val.lower() == "true":
            return True
        if val.lower() == "false":
            return False

        return val

    @property
    def all(self):
        return self._sections

    def get(self, key, section=None, expand_to_list=False):
        """ If section is None use flattened """
        if section is None:
            value = self._flattened_config.get(key)
        else:
            if section not in self._sections:
                log.debug("section '%s' not found in config file, "
                          "trying lower case", section)
                section = section.lower()

            value = self._sections.get(section, {}).get(key)

        if expand_to_list:
            return self.expand_value_ranges(value)

        return value

    @property
    def dump(self):
        with open(self.path) as fd:
            return fd.read()

    def _load(self):
        if not self.exists:
            return

        current_section = None
        with open(self.path) as fd:
            for line in fd:
                if re.compile(r"^\s*#").search(line):
                    continue

                # section names are not expected to contain whitespace
                ret = re.compile(r"^\s*\[(\S+)].*").search(line)
                if ret:
                    current_section = ret.group(1)
                    self._sections[current_section] = {}
                    continue

                if current_section is None:
                    continue

                # key names may contain whitespace
                # values may contain whitespace
                expr = r"^\s*(\S+(?:\s+\S+)?)\s*=\s*(.+)\s*"
                ret = re.compile(expr).search(line)
                if ret:
                    key = ret.group(1)
                    val = self.bool_str(ret.group(2))
                    if type(val) == str:
                        val = val.strip()
                        for char in ["'", '"']:
                            val = val.strip(char)

                    self._sections[current_section][key] = val
                    self._flattened_config[key] = val
